function [conn, ind, cpos, fconn] = coarseConnections(g, p, varargin)
%Derive connection/topology structure on coarse grid
%
% SYNOPSIS:
%    conn                    = coarseConnections(G, p)
%    conn                    = coarseConnections(G, p, Ic)
%   [conn, ind, cpos, fconn] = coarseConnections(...)
%
% PARAMETERS:
%   G  - Fine-scale grid structure over which to derive coarse topology.
%
%   p  - Partition vector for fine-scale entities (cells).
%
%   Ic - Discrete indicator function.  An m-by-n numeric array of function
%        values.  There must be one row in 'Ic' for each connection (row)
%        in the neighbourship definition (in other words m==SIZE(N,1)).
%        OPTIONAL.  Default value: Ic = ONES([m, 1]).
%
% RETURNS:
%   conn - Coarse-scale connection structure represented as a neighbourship
%          definition (m-by-2 array of coarse-scale block pairs).  A
%          coarse-scale connection is defined as the intersection of the
%          unique block pairs defined by p(N) with the unique applicable
%          indicator tuples defined by 'Ic(c,:)'.
%
%   ind  - Actual indicator value (including half-face tag if present)
%          associated with each coarse-scale connection.
%
%   cpos, fconn -
%          Packed data-array representation of coarse->fine connection
%          mapping.  Specifically, the elements
%
%              fconn(cpos(i) : cpos(i + 1) - 1)
%
%          are the fine-scale connections (i.e., rows of the neighbourship
%          definition 'N') that constitute coarse-scale connection 'i'.
%
% NOTE:
%   This function uses SORTROWS.
%
% SEE ALSO:
%   partitionCartGrid, generateCoarseGrid, sortrows.

%{
Copyright 2009, 2010, 2011, 2012 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Date: 2012-09-11 15:32:18 +0200 (Tue, 11 Sep 2012) $
% $Revision: 9662 $

   N = g.faces.neighbors;
   if (nargin > 2) && isnumeric(varargin{1}),
      Ic = varargin{1};
   else
      Ic = ones([size(N,1), 1]);
   end

   assert (size(Ic,1) == size(N,1), ...
           'Connection indicators must be defined for each connection');

   % Enumerate connections.
   p = [0; p];
   c = (1 : size(N,1)) .';

   pN = p(N + 1);
   i  = pN(:,1) ~= pN(:,2);

   pN = sort(double(pN(i,:)), 2);
   c  = c(i);
   Ic = double(Ic(i,:));

   if false && size(g.nodes.coords, 2) == 2,
      fn     = g.faces.nodes(rldecode(i,2));
      w      = size([pN, Ic, c],2);
      A      = sortrows([rldecode([pN, Ic, c], 2), fn], [1:w-1,w+1,w]);
      [n,n]  = rlencode(A(:,[1,2,3,4,end]));
      n      = rldecode(n,n);
      cpos   = find(n([1:2:end,end]) == 1);
      
      A = A(1:2:end, :);
      fconn  = A(:,end-1);
      
      %[B, n] = rlencode([A(:, 1 : end-2));
      conn   = A(cpos(1:end-1),1:2);
   else
      A      = sortrows([pN, Ic, c]);
      [B, n] = rlencode(A(:, 1 : end-1));
      conn  = B(:,1:2);
      ind   = B(:,3:end);
      cpos  = cumsum([1; reshape(n, [], 1)]);
      fconn = A(:, end);
   end
end
