function CG = generateCoarseGrid(G, partition, varargin)
%Build coarse grid data structure from partion of existing fine grid.
%
% SYNOPSIS:
%   CG = generateCoarseGrid(G, p)
%
% PARAMETERS:
%   G - grid_structure data structure describing fine-scale discretisation
%       of reservoir geometry.
%
%   p - Partition vector of size [G.cells.num, 1] describing the coarse
%       grid.  We assume that all coarse blocks are connected.  The
%       partition vector is often created by function partitionCartGrid
%       or function partitionUI.
%
% RETURNS:
%   CG - Coarse grid structure.  The coarse grid consists entirely of
%        topological information stored in the same topological fields as
%        in the fine-scale grid described in 'grid_structure'.
%
%        Specifically, the fields
%
%            CG.cells.num, CG.cells.facePos, CG.cells.faces
%            CG.faces.num, and CG.faces.neighbors
%
%        have the same interpretation in the coarse grid as in the
%        fine-scale grid.
%
%        The partition vector 'p' is stored within the coarse grid in a
%        separate field, 'CG.partition'.
%
% NOTES:
%   The coarse grid does not define nodes so the fine-scale fields
%   'faces.nodePos' and 'faces.nodes' are not defined.
%
%   Face areas and block volumes must be calculated separately.
%
% SEE ALSO:
%   grid_structure, generateCoarseSystem, generateCoarseWellSystem.

%{
Copyright 2009, 2010, 2011, 2012 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Date: 2012-09-11 15:32:18 +0200 (Tue, 11 Sep 2012) $
% $Revision: 9662 $

   %-----------------------------------------------------------------------
   %- Re-index coarse blocks according to pos volume ----------------------
   %
   [partition, numCC] = get_active(reshape(partition, [], 1));

   %-----------------------------------------------------------------------
   %- Detect neighbouring coarse blocks -----------------------------------
   %
   [i, j] = connection(G, partition);

   %-----------------------------------------------------------------------
   %- Find coarse boundary faces ------------------------------------------
   %
   [ctags, ibdry, jbdry] = boundary_faces(G, partition);

   % Tag all internal coarse faces with tag zero.
   ctags = vertcat(zeros([numel(j), 1]), ctags{:});

   % Add coarse boundary faces to neighbour matrix.
   i = [i; ibdry];
   j = [j; jbdry];

   %-----------------------------------------------------------------------
   %- Build final coarse grid structure -----------------------------------
   %
   CG.cells.num       = numCC;
   CG.cells.faces     = int32(cell_faces(i, j));
   CG.cells.facePos   = find(diff([0; CG.cells.faces(:,end); CG.cells.num+1]));

   CG.faces.num       = numel(i);
   CG.faces.neighbors = int32([i, j]);
   CG.faces.tag       = ctags;

   CG.partition       = partition;
   CG.griddim         = G.griddim;
end

%--------------------------------------------------------------------------
% Helpers follow.
%--------------------------------------------------------------------------

function [p, nb] = get_active(p)
   nb = max(p);

   B    = find(accumarray(p, 1));
   I    = zeros([nb, 1]);
   I(B) = 1 : numel(B);

   p  = I(p);
   nb = max(p);      % Number of coarse blocks.
end

%--------------------------------------------------------------------------

function [i, j] = connection(g, p)
   p = [1; reshape(p, [], 1) + 1];
   N = p(g.faces.neighbors + 1);
   N = N(N(:,1) ~= N(:,2), :);  % Exclude block-internal connections.

   % Build (symmetric) connectivity matrix.
   conn = sparse([N(:,1) ; N(:,2)], ...
                 [N(:,2) ; N(:,1)],  1 );

   % Extract coarse block neighbours to be entered into first ('i') and
   % second ('j') column of array 'CG.faces.neighbors'.
   [j, i] = find(tril(conn(2:end, 2:end), -1));
end

%--------------------------------------------------------------------------

function [ctags, i, j] = boundary_faces(g, p)
   % Extract (fine-scale) tags from all exterior (fine-scale) faces and the
   % coarse blocks to which these fine-scale faces are connected.
   %
   ext        = any(g.faces.neighbors == 0, 2);
   block      = zeros([g.faces.num, 1]);
   block(ext) = p(sum(g.faces.neighbors(ext,:), 2));

   f = g.cells.faces(ext(g.cells.faces(:,1)), :);

   % Determine tags on exterior coarse faces.
   %
   %  1) Extract unique (fine-scale) tags for all coarse blocks in 'block'.
   %     Result 'ctags' is a cell array of tags (see ACCUMARRAY).
   %
   ctags = accumarray(block(f(:,1)), f(:,2),  ...
                      [max(block(f(:,1))), 1], @(t) { unique(t) });

   %  2) Keep only the blocks for which any tags are assigned
   %     (i.e., the boundary blocks (stored in 'keep')).
   %
   keep  = cellfun(@(t) ~isempty(t), ctags);
   ctags = ctags(keep);

   % Count number of additional basis functions (i.e., faces) on boundary
   % blocks.
   %
   numBF = cellfun(@(t) numel(t), ctags);

   % Create cell-neighbour entries for boundary faces.  The blocks in
   % 'keep' connect to the outside (block zero).  Array 'i' will be entered
   % into the first column of 'CG.faces.neighbors' while array 'j' will be
   % entered into the second column.
   %
   i = zeros   ([sum(numBF),     1]);
   j = rldecode( find(keep), numBF );
end

%--------------------------------------------------------------------------

function cf = cell_faces(i, j)
   nn = numel(i);

   %-----------------------------------------------------------------------
   % List local faces and corresponding global index ----------------------
   %
   V = sortrows([ [i, (1 : nn).'];
                  [j, (1 : nn).'] ]);

   i1 = find(V(:,1), 1, 'first'); % First non-zero cell (block).

   cf = V(i1:end, end : -1 : 1);
end
